# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import TypeVar, Callable, Type, Any

T = TypeVar("T")

def proto_base_name(func: Callable[..., Any]) -> str:
    return "".join([x.capitalize() for x in func.__qualname__.split(".")])

def proto_function_service_name(func: Callable[..., Any]) -> str:
    return f"{proto_base_name(func)}FunctionService"

def proto_function_rpc_name(func: Callable[..., Any]) -> str:
    return proto_base_name(func)

def proto_function_req_name(func: Callable[..., Any]) -> str:
    return f"{proto_base_name(func)}FunctionRequest"

def proto_function_res_name(func: Callable[..., Any]) -> str:
    return f"{proto_base_name(func)}FunctionResponse"

def proto_dataclass_name(cls: Type[T]) -> str:
    return f"{proto_base_name(cls)}DataClassMessage"

def proto_remoteclass_service_name(cls: Type[T]) -> str:
    return f"{proto_base_name(cls)}RemoteClassService"

def proto_method_rpc_name(cls: Type[T], method: Callable[..., Any]) -> str:
    cls_name = proto_base_name(cls)

    if method.__name__ == "__init__":
        return f"{cls_name}Init"
    elif method.__name__ == "__del__":
        return f"{cls_name}Del"
    else:
        base = proto_base_name(method)
        method_name = base[len(cls_name):] if base.startswith(cls_name) else base
        return f"{cls_name}{method_name}Method"

def proto_method_req_name(cls: Type[T], method: Callable[..., Any]) -> str:
    cls_name = proto_base_name(cls)

    if method.__name__ == "__init__":
        return f"{cls_name}InitRemoteClassRequest"
    elif method.__name__ == "__del__":
        return f"{cls_name}DelRemoteClassRequest"
    else:
        base = proto_base_name(method)
        method_name = base[len(cls_name):] if base.startswith(cls_name) else base
        return f"{cls_name}{method_name}MethodRemoteClassRequest"

def proto_method_res_name(cls: Type[T], method: Callable[..., Any]) -> str:
    cls_name = proto_base_name(cls)

    if method.__name__ == "__init__":
        return f"{cls_name}InitRemoteClassResponse"
    elif method.__name__ == "__del__":
        return f"{cls_name}DelRemoteClassResponse"
    else:
        base = proto_base_name(method)
        method_name = base[len(cls_name):] if base.startswith(cls_name) else base
        return f"{cls_name}{method_name}MethodRemoteClassResponse"

def proto_instance_ref_name(cls: Type[T]) -> str:
    return f"{proto_base_name(cls)}InstanceMessage"



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import get_type_hints, get_args, Union, List, Tuple, Dict, TypeVar, Callable, Any, Type, Optional
from typing_extensions import ParamSpec
from types import ModuleType
from collections.abc import Sequence, Mapping

from . import names
from .util import is_type_match, get_function_typed_params, get_function_return_type, get_method_typed_params, get_method_return_type, normalize_args, TypeHint, my_get_origin
from .compiler import proto_primitive_type_mapping, proto_dataclass_type_mapping, proto_remoteclass_type_mapping, proto_remoteclass_id_mapping, proto_remoteclass_id_reverse_mapping, is_subclass, subclasses, EllipsisType
from .instance_ref import InstanceRefType, get_ref_from_instance, get_instance_from_ref
from .registry import get_handler_for_type

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")

ProtoMsg = Any

dynamic_pb2: Optional[ModuleType] = None

def init_proto(module: ModuleType) -> None:
    global dynamic_pb2
    dynamic_pb2 = module

def get_proto_module() -> ModuleType:
    global dynamic_pb2
    assert dynamic_pb2 is not None
    return dynamic_pb2

def new_proto_function_request(func: Callable[P, R]) -> ProtoMsg:
    return getattr(get_proto_module(), names.proto_function_req_name(func))()

def new_proto_function_response(func: Callable[P, R]) -> ProtoMsg:
    return getattr(get_proto_module(), names.proto_function_res_name(func))()

def pack_proto_function_request(func: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> ProtoMsg:
    typed_params = get_function_typed_params(func)
    normalized_args = normalize_args(func, *args, **kwargs)

    msg = new_proto_function_request(func)

    for param_name, arg in normalized_args.items():
        set_proto_field(msg, param_name, typed_params[param_name], arg, allow_subclass=True, on_server=False)

    return msg

def unpack_proto_function_request(func: Callable[P, R], msg: ProtoMsg) -> Dict[str, Any]:
    typed_params = get_function_typed_params(func)
    return {param_name: get_proto_field(msg, param_name, type_hint, allow_subclass=True, on_server=True) \
            for param_name, type_hint in typed_params.items()}

def pack_proto_function_response(func: Callable[P, R], obj: R) -> ProtoMsg:
    msg = new_proto_function_response(func)
    return_type = get_function_return_type(func)
    set_proto_field(msg, "return", return_type, obj, allow_subclass=True, on_server=True)
    return msg

def unpack_proto_function_response(func: Callable[P, R], msg: ProtoMsg) -> Any:
    return_type = get_function_return_type(func)
    return get_proto_field(msg, "return", return_type, allow_subclass=True, on_server=False)

def new_proto_method_request(cls: Type[T], method: Callable[P, R]) -> ProtoMsg:
    return getattr(get_proto_module(), names.proto_method_req_name(cls, method))()

def new_proto_method_response(cls: Type[T], method: Callable[P, R]) -> ProtoMsg:
    return getattr(get_proto_module(), names.proto_method_res_name(cls, method))()

def pack_proto_method_request(cls: Type[T], method: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> ProtoMsg:
    typed_params = get_method_typed_params(cls, method)
    normalized_args = normalize_args(method, *args, **kwargs)

    msg = new_proto_method_request(cls, method)

    for i, (param_name, arg) in enumerate(normalized_args.items()):
        if i == 0 and method.__name__ == "__init__":
            continue
        set_proto_field(msg, param_name, typed_params[param_name], arg, allow_subclass=True, on_server=False)

    return msg

def unpack_proto_method_request(cls: Type[T], method: Callable[P, R], msg: ProtoMsg) -> Dict[str, Any]:
    typed_params = get_method_typed_params(cls, method)

    if method.__name__ == "__init__":
        typed_params = dict(list(typed_params.items())[1:])

    return {param_name: get_proto_field(msg, param_name, type_hint, allow_subclass=True, on_server=True) \
            for param_name, type_hint in typed_params.items()}

def pack_proto_method_response(cls: Type[T], method: Callable[P, R], obj: R) -> ProtoMsg:
    msg = new_proto_method_response(cls, method)

    if method.__name__ == "__init__":
        instance_ref = get_ref_from_instance(cls, obj, True)
        set_proto_field(msg, "return", InstanceRefType, instance_ref, allow_subclass=True, on_server=True)
        return msg

    elif method.__name__ == "__del__":
        return msg

    else:
        return_type = get_method_return_type(cls, method)
        set_proto_field(msg, "return", return_type, obj, allow_subclass=True, on_server=True)
        return msg

def unpack_proto_method_response(cls: Type[T], method: Callable[P, R], msg: ProtoMsg) -> Any:
    if method.__name__ == "__init__":
        return get_proto_field(msg, "return", InstanceRefType, allow_subclass=True, on_server=False)

    elif method.__name__ == "__del__":
        return None

    else:
        return_type = get_method_return_type(cls, method)
        o = get_proto_field(msg, "return", return_type, allow_subclass=True, on_server=False)
        return o

def get_proto_field(proto_msg: ProtoMsg, param_name: str, type_hint: TypeHint, allow_subclass: bool, on_server: bool) -> Any:
    type_origin = my_get_origin(type_hint)
    type_args = get_args(type_hint)

    type_origin = type_hint if type_origin is None else type_origin

    if type_origin is type(None):
        return None

    elif type_origin is EllipsisType:
        return ...

    elif type_origin in proto_primitive_type_mapping:
        return getattr(proto_msg, param_name)

    elif type_origin is slice:
        msg = getattr(proto_msg, param_name)
        start = msg.start if msg.HasField("start") else None
        stop = msg.stop if msg.HasField("stop") else None
        step = msg.step if msg.HasField("step") else None
        return slice(start, stop, step)

    elif is_subclass(type_origin, proto_dataclass_type_mapping):
        proto_class_msg = getattr(proto_msg, param_name)
        proto_types = subclasses(type_origin, proto_dataclass_type_mapping)
        if not allow_subclass or len(proto_types) == 0:
            args = {member_name: get_proto_field(proto_class_msg, member_name, th, allow_subclass=True, on_server=on_server)
                    for member_name, th in get_type_hints(type_origin).items()}
            return type_origin(**args)
        else:
            for i, th in enumerate(proto_types.keys()):
                if proto_class_msg.HasField(f"class{i}"):
                    return get_proto_field(proto_class_msg, f"class{i}", th, allow_subclass=False, on_server=on_server)
            raise ValueError(f"Parameter '{param_name}' is empty.")

    elif type_origin in proto_remoteclass_type_mapping:
        cls = type_origin
        msg = getattr(proto_msg, param_name)
        instance_ref = msg.ref

        actual_cls = cls
        if allow_subclass:
            candidate_cls = proto_remoteclass_id_reverse_mapping.get(msg.type_id)
            if candidate_cls is not None and issubclass(candidate_cls, cls):
                actual_cls = candidate_cls

        return get_instance_from_ref(actual_cls, instance_ref, type_hint, on_server)

    elif type_origin is Union:
        child_proto_msg = getattr(proto_msg, param_name)
        for i, th in enumerate(type_args):
            if child_proto_msg.HasField(f"member{i}"):
                return get_proto_field(child_proto_msg, f"member{i}", th, allow_subclass=allow_subclass, on_server=on_server)
        raise ValueError(f"Parameter '{param_name}' is empty.")

    elif type_origin in (tuple, Tuple):
        if len(type_args) == 2 and type_args[1] is Ellipsis:
            repeated_container = getattr(proto_msg, param_name).items
            return tuple(get_proto_repeated_field(repeated_container, param_name, type_args[0], allow_subclass=allow_subclass, on_server=on_server))
        else:
            child_proto_msg = getattr(proto_msg, param_name)
            objs = []
            for i, th in enumerate(type_args):
                objs.append(get_proto_field(child_proto_msg, f"item{i}", th, allow_subclass=allow_subclass, on_server=on_server))
            return tuple(objs)

    elif type_origin in (list, List, Sequence):
        repeated_container = getattr(proto_msg, param_name).elements
        return get_proto_repeated_field(repeated_container, param_name, type_args[0], allow_subclass=type_origin is Sequence, on_server=on_server)

    elif type_origin in (dict, Dict, Mapping):
        repeated_container_k = getattr(proto_msg, param_name).keys
        repeated_container_v = getattr(proto_msg, param_name).values
        # Mapping key type is not covariant
        keys = get_proto_repeated_field(repeated_container_k, param_name, type_args[0], allow_subclass=False, on_server=on_server)
        values = get_proto_repeated_field(repeated_container_v, param_name, type_args[1], allow_subclass=type_origin is Mapping, on_server=on_server)
        assert len(keys) == len(values)
        return dict(zip(keys, values))

    else:
        handler = get_handler_for_type(type_origin)
        if handler is not None:
            payload = get_proto_field(proto_msg, param_name, handler, allow_subclass=allow_subclass, on_server=on_server)
            return payload.unpack()
        else:
            raise TypeError(f"Type {type_origin} is not supported.")

def get_proto_repeated_field(repeated_container: Any, param_name: str, type_hint: TypeHint, allow_subclass: bool, on_server: bool) -> List[Any]:
    if type_hint in proto_primitive_type_mapping:
        # RepeatedScalarContainer
        return list(repeated_container)
    else:
        # RepeatedCompositeContainer
        objs = []
        for child_proto_msg in repeated_container:
            WrapperMsg = type("WrapperMessage", (), {param_name: child_proto_msg})
            objs.append(get_proto_field(WrapperMsg, param_name, type_hint, allow_subclass=allow_subclass, on_server=on_server))
        return objs

def set_proto_field(proto_msg: ProtoMsg, param_name: str, type_hint: TypeHint, obj: Any, allow_subclass: bool, on_server: bool) -> None:
    type_origin = my_get_origin(type_hint)
    type_args = get_args(type_hint)

    type_origin = type_hint if type_origin is None else type_origin

    if type_origin is type(None):
        setattr(proto_msg, param_name, True)

    elif type_origin is EllipsisType:
        setattr(proto_msg, param_name, True)

    elif type_origin in proto_primitive_type_mapping:
        setattr(proto_msg, param_name, obj)

    elif type_origin is slice:
        msg = getattr(proto_msg, param_name)
        msg.is_present = True
        if obj.start is not None:
            msg.start = obj.start
        if obj.stop is not None:
            msg.stop = obj.stop
        if obj.step is not None:
            msg.step = obj.step

    elif is_subclass(type_origin, proto_dataclass_type_mapping):
        proto_class_msg = getattr(proto_msg, param_name)
        proto_types = subclasses(type_origin, proto_dataclass_type_mapping)
        if not allow_subclass or len(proto_types) == 0:
            for member_name, th in get_type_hints(type_origin).items():
                set_proto_field(proto_class_msg, member_name, th, getattr(obj, member_name), allow_subclass=True, on_server=on_server)
        else:
            for i, th in enumerate(proto_types.keys()):
                if type(obj) is th:
                    set_proto_field(proto_class_msg, f"class{i}", th, obj, allow_subclass=False, on_server=on_server)
                    return
            raise TypeError(f"Type '{type(obj)}' of parameter '{param_name}' does not match any of {type_hint}.")

    elif type_origin in proto_remoteclass_type_mapping:
        cls = type_origin
        actual_cls = type(obj) if allow_subclass and isinstance(obj, cls) else cls

        if actual_cls not in proto_remoteclass_id_mapping:
            raise TypeError(f"Class '{actual_cls.__name__}' is not registered as @egrpc.remoteclass. "
                            f"All subclasses of '{cls.__name__}' must be decorated with @egrpc.remoteclass.")

        instance_ref = get_ref_from_instance(actual_cls, obj, on_server)
        msg = getattr(proto_msg, param_name)
        msg.ref = instance_ref
        msg.type_id = proto_remoteclass_id_mapping[actual_cls]

    elif type_origin is Union:
        child_proto_msg = getattr(proto_msg, param_name)
        for i, th in enumerate(type_args):
            # TODO: consider what to do when obj matches multiple candidates
            if is_type_match(obj, th):
                set_proto_field(child_proto_msg, f"member{i}", th, obj, allow_subclass=allow_subclass, on_server=on_server)
                return
        raise TypeError(f"Type '{type(obj)}' of parameter '{param_name}' does not match any of {type_hint}.")

    elif type_origin in (tuple, Tuple):
        if len(type_args) == 2 and type_args[1] is Ellipsis:
            repeated_container = getattr(proto_msg, param_name).items
            set_proto_repeated_field(repeated_container, param_name, type_args[0], obj, allow_subclass=allow_subclass, on_server=on_server)
        else:
            child_proto_msg = getattr(proto_msg, param_name)
            for i, (th, o) in enumerate(zip(type_args, obj)):
                set_proto_field(child_proto_msg, f"item{i}", th, o, allow_subclass=allow_subclass, on_server=on_server)

    elif type_origin in (list, List, Sequence):
        repeated_container = getattr(proto_msg, param_name).elements
        set_proto_repeated_field(repeated_container, param_name, type_args[0], obj, allow_subclass=type_origin is Sequence, on_server=on_server)

    elif type_origin in (dict, Dict, Mapping):
        repeated_container_k = getattr(proto_msg, param_name).keys
        repeated_container_v = getattr(proto_msg, param_name).values
        # Mapping key type is not covariant
        set_proto_repeated_field(repeated_container_k, param_name, type_args[0], obj.keys(), allow_subclass=False, on_server=on_server)
        set_proto_repeated_field(repeated_container_v, param_name, type_args[1], obj.values(), allow_subclass=type_origin is Mapping, on_server=on_server)

    else:
        handler = get_handler_for_type(type_origin)
        if handler is not None:
            payload_obj = handler.pack(obj)
            set_proto_field(proto_msg, param_name, handler, payload_obj, allow_subclass=allow_subclass, on_server=on_server)
        else:
            raise TypeError(f"Type {type_origin} is not supported.")

def set_proto_repeated_field(repeated_container: Any, param_name: str, type_hint: TypeHint, objs: List[Any], allow_subclass: bool, on_server: bool) -> None:
    if type_hint in proto_primitive_type_mapping:
        # RepeatedScalarContainer
        repeated_container.extend(objs)
    else:
        # RepeatedCompositeContainer
        for o in objs:
            # https://stackoverflow.com/questions/57222346/how-to-get-type-contained-by-protobufs-repeatedcompositecontainer-or-repeatedsc
            child_proto_msg = repeated_container.add()
            WrapperMsg = type("WrapperMessage", (), {param_name: child_proto_msg})
            set_proto_field(WrapperMsg, param_name, type_hint, o, allow_subclass=allow_subclass, on_server=on_server)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import get_args, Union, List, Tuple, Dict, Callable, Any, TypeVar, Type, Optional
from typing_extensions import ParamSpec
from types import ModuleType
from collections.abc import Sequence, Mapping
import os
import sys
import tempfile
import importlib.util
from grpc_tools import protoc # type: ignore[import-untyped]

from . import names
from .util import get_function_typed_params, get_function_return_type, get_class_typed_members, get_method_typed_params, get_method_return_type, TypeHint, my_get_origin
from .instance_ref import InstanceRefType
from .registry import get_handler_for_type, get_primitive_type_registry

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")

EllipsisType = type(...)

proto_primitive_type_mapping = {
    str          : "string",
    int          : "int64",
    float        : "double",
    bool         : "bool",
    type(None)   : "bool",
    EllipsisType : "bool",
    bytes        : "bytes",
}

proto_dataclass_type_mapping: Dict[Any, str] = {}
proto_remoteclass_type_mapping: Dict[Any, str] = {}
proto_remoteclass_id_mapping: Dict[Any, int] = {}
proto_remoteclass_id_reverse_mapping: Dict[int, Any] = {}

proto_header = """syntax = "proto3";

message SliceSpec {
  optional int64 start = 1;
  optional int64 stop = 2;
  optional int64 step = 3;
  bool is_present = 4;
}
"""

proto_content = ""

def indent_str(depth: int) -> str:
    return " " * depth * 2

def is_subclass(type_hint: TypeHint, type_mapping: Dict[Any, str]) -> bool:
    return isinstance(type_hint, type) and issubclass(type_hint, tuple(type_mapping.keys()))

def subclasses(type_hint: TypeHint, type_mapping: Dict[Any, str]) -> Dict[TypeHint, str]:
    if isinstance(type_hint, type):
        return {th: proto_type for th, proto_type in type_mapping.items() if issubclass(th, type_hint)}
    else:
        return {}

def resolve_custom_type(type_hint: TypeHint) -> TypeHint:
    type_origin = my_get_origin(type_hint)

    if isinstance(type_hint, type):
        handler = get_handler_for_type(type_hint)
    elif isinstance(type_origin, type):
        handler = get_handler_for_type(type_origin)
    else:
        return type_hint

    if handler is not None:
        return handler
    else:
        return type_hint

def gen_proto_field_def(index          : int,
                        param_name     : str,
                        type_hint      : TypeHint,
                        allow_subclass : bool = True,
                        repeated       : bool = False,
                        depth          : int = 0) -> Tuple[List[str], List[str], int]:
    type_hint = resolve_custom_type(type_hint)
    type_origin = my_get_origin(type_hint)
    type_args = get_args(type_hint)

    type_origin = type_hint if type_origin is None else type_origin

    repeated_str = "repeated " if repeated else ""

    proto_fields = []
    proto_defs = []

    if type_origin in proto_primitive_type_mapping:
        proto_type = proto_primitive_type_mapping[type_origin]
        proto_fields.append(f"{indent_str(depth)}{repeated_str}{proto_type} {param_name} = {index + 1};")
        index += 1

    elif type_origin is slice:
        proto_fields.append(f"{indent_str(depth)}{repeated_str}SliceSpec {param_name} = {index + 1};")
        index += 1

    elif is_subclass(type_origin, proto_dataclass_type_mapping):
        # TODO: move subtype handling to `compile_dataclass()`?
        proto_types = subclasses(type_origin, proto_dataclass_type_mapping)
        if not allow_subclass or len(proto_types) == 0:
            proto_type = proto_dataclass_type_mapping[type_origin]
            proto_fields.append(f"{indent_str(depth)}{repeated_str}{proto_type} {param_name} = {index + 1};")
            index += 1
        else:
            msgname = f"{param_name.capitalize()}SubclassMessage"
            child_type_hints = {f"class{i}": th for i, th in enumerate(proto_types.keys())}
            proto_defs += gen_proto_msg_def(msgname, child_type_hints, allow_subclass=False, oneof=True)
            proto_fields.append(f"{indent_str(depth)}{repeated_str}{msgname} {param_name} = {index + 1};")
            index += 1

    elif type_origin in proto_remoteclass_type_mapping:
        proto_type = proto_remoteclass_type_mapping[type_origin]
        proto_fields.append(f"{indent_str(depth)}{repeated_str}{proto_type} {param_name} = {index + 1};")
        index += 1

    elif type_origin is Union:
        msgname = f"{param_name.capitalize()}UnionMessage"
        child_type_hints = {f"member{i}": th for i, th in enumerate(type_args)}
        proto_defs += gen_proto_msg_def(msgname, child_type_hints, allow_subclass=allow_subclass, oneof=True)
        proto_fields.append(f"{indent_str(depth)}{repeated_str}{msgname} {param_name} = {index + 1};")
        index += 1

    elif type_origin in (tuple, Tuple):
        if len(type_args) == 2 and type_args[1] is Ellipsis:
            msgname = f"{param_name.capitalize()}TupleVMessage"
            proto_defs += gen_proto_msg_def(msgname, {"items": type_args[0]}, allow_subclass=allow_subclass, repeated=True)
        else:
            msgname = f"{param_name.capitalize()}TupleMessage"
            child_type_hints = {f"item{i}": th for i, th in enumerate(type_args)}
            proto_defs += gen_proto_msg_def(msgname, child_type_hints, allow_subclass=allow_subclass)
        proto_fields.append(f"{indent_str(depth)}{repeated_str}{msgname} {param_name} = {index + 1};")
        index += 1

    elif type_origin in (list, List, Sequence):
        msgname = f"{param_name.capitalize()}ListMessage"
        proto_defs += gen_proto_msg_def(msgname, {"elements": type_args[0]}, allow_subclass=type_origin is Sequence, repeated=True)
        proto_fields.append(f"{indent_str(depth)}{repeated_str}{msgname} {param_name} = {index + 1};")
        index += 1

    elif type_origin in (dict, Dict, Mapping):
        msgname = f"{param_name.capitalize()}DictMessage"
        proto_defs += gen_proto_msg_def(msgname, {"keys": type_args[0], "values": type_args[1]}, allow_subclass=type_origin is Mapping, repeated=True)
        proto_fields.append(f"{indent_str(depth)}{repeated_str}{msgname} {param_name} = {index + 1};")
        index += 1

    else:
        raise TypeError(f"Type {type_origin} is not supported.")

    return proto_fields, proto_defs, index

def gen_proto_msg_def(msgname        : str,
                      typed_params   : Dict[str, TypeHint],
                      allow_subclass : bool = True,
                      repeated       : bool = False,
                      oneof          : bool = False,
                      depth          : int = 0) -> List[str]:
    index = 0
    proto_defs = []
    proto_inner_defs = []

    proto_defs.append(f"{indent_str(depth)}message {msgname} {{")

    if oneof:
        proto_defs.append(f"{indent_str(depth + 1)}oneof wrapper {{")

    for param_name, type_hint in typed_params.items():
        next_depth = depth + 2 if oneof else depth + 1
        pf, pd, index = gen_proto_field_def(index, param_name, type_hint, allow_subclass=allow_subclass, repeated=repeated, depth=next_depth)
        proto_defs += pf
        proto_inner_defs += pd

    if oneof:
        proto_defs.append(f"{indent_str(depth + 1)}}}")

    proto_defs += [indent_str(depth + 1) + line for line in proto_inner_defs]

    proto_defs.append(f"{indent_str(depth)}}}")

    return proto_defs

deferred_compilation: List[Callable[[], None]] = []

def defer(func: Callable[[], None]) -> None:
    global deferred_compilation
    deferred_compilation.append(func)

def do_deferred() -> None:
    global deferred_compilation
    pending = deferred_compilation
    deferred_compilation = []

    last_error: Optional[Exception] = None

    while pending:
        next_pending = []
        progress = False

        for func in pending:
            try:
                func()
            except Exception as exc:
                next_pending.append(func)
                last_error = exc
            else:
                progress = True

        if not progress:
            assert last_error
            raise last_error

        pending = next_pending

def compile_function(func: Callable[P, R]) -> None:
    def do_compile() -> None:
        typed_params = get_function_typed_params(func)
        return_type = get_function_return_type(func)

        proto_service_name = names.proto_function_service_name(func)
        proto_rpc_name = names.proto_function_rpc_name(func)
        proto_req_name = names.proto_function_req_name(func)
        proto_res_name = names.proto_function_res_name(func)

        proto_req_def = gen_proto_msg_def(proto_req_name, typed_params)
        proto_res_def = gen_proto_msg_def(proto_res_name, {"return": return_type})

        global proto_content
        proto_content += f"""
service {proto_service_name} {{
  rpc {proto_rpc_name} ({proto_req_name}) returns ({proto_res_name});
}}

{chr(10).join(proto_req_def)}

{chr(10).join(proto_res_def)}
"""

    defer(do_compile)

def compile_dataclass(cls: Type[T]) -> None:
    def do_compile() -> None:
        proto_msgname = names.proto_dataclass_name(cls)

        # Register beforehand so recursive references can resolve
        global proto_dataclass_type_mapping
        proto_dataclass_type_mapping[cls] = proto_msgname

        proto_def = gen_proto_msg_def(proto_msgname, get_class_typed_members(cls))

        global proto_content
        proto_content += "\n"
        proto_content += "\n".join(proto_def)
        proto_content += "\n"

    defer(do_compile)

proto_remoteclass_rpc_defs: Dict[Type[Any], List[str]] = {}
proto_remoteclass_msg_defs: Dict[Type[Any], List[str]] = {}

def add_remoteclass_method_def(cls: Type[T], proto_rpc_def: str, proto_req_def: List[str], proto_res_def: List[str]) -> None:
    global proto_remoteclass_rpc_defs, proto_remoteclass_msg_defs

    if cls not in proto_remoteclass_rpc_defs:
        proto_remoteclass_rpc_defs[cls] = []
    if cls not in proto_remoteclass_msg_defs:
        proto_remoteclass_msg_defs[cls] = []

    proto_remoteclass_rpc_defs[cls].append(proto_rpc_def)
    proto_remoteclass_msg_defs[cls] += [""] + proto_req_def
    proto_remoteclass_msg_defs[cls] += [""] + proto_res_def

def define_remoteclass_instance_ref(cls: Type[T]) -> None:
    global proto_remoteclass_type_mapping
    global proto_remoteclass_id_mapping
    global proto_remoteclass_id_reverse_mapping

    if cls not in proto_remoteclass_type_mapping:
        proto_instance_ref_name = names.proto_instance_ref_name(cls)

        proto_remoteclass_type_mapping[cls] = proto_instance_ref_name
        type_id = len(proto_remoteclass_id_mapping)
        proto_remoteclass_id_mapping[cls] = type_id
        proto_remoteclass_id_reverse_mapping[type_id] = cls

        global proto_remoteclass_msg_defs

        if cls not in proto_remoteclass_msg_defs:
            proto_remoteclass_msg_defs[cls] = []

        proto_instance_def = gen_proto_msg_def(
            proto_instance_ref_name,
            {
                "ref": InstanceRefType,
                "type_id": int,
            },
        )
        proto_remoteclass_msg_defs[cls] += [""] + proto_instance_def

def compile_remoteclass(cls: Type[T]) -> None:
    define_remoteclass_instance_ref(cls)

    def do_compile() -> None:
        proto_service_name = names.proto_remoteclass_service_name(cls)

        global proto_content
        proto_content += f"""
service {proto_service_name} {{
{chr(10).join(proto_remoteclass_rpc_defs[cls])}
}}
{chr(10).join(proto_remoteclass_msg_defs[cls])}
"""

        del proto_remoteclass_rpc_defs[cls]
        del proto_remoteclass_msg_defs[cls]

    defer(do_compile)

def compile_remoteclass_init(cls: Type[T], init_method: Callable[P, R]) -> None:
    define_remoteclass_instance_ref(cls)

    def do_compile() -> None:
        typed_params = get_method_typed_params(cls, init_method)

        proto_rpc_name = names.proto_method_rpc_name(cls, init_method)
        proto_req_name = names.proto_method_req_name(cls, init_method)
        proto_res_name = names.proto_method_res_name(cls, init_method)

        proto_rpc_def = f"  rpc {proto_rpc_name} ({proto_req_name}) returns ({proto_res_name});"
        proto_req_def = gen_proto_msg_def(proto_req_name, dict(list(typed_params.items())[1:]))
        proto_res_def = gen_proto_msg_def(proto_res_name, {"return": InstanceRefType})

        add_remoteclass_method_def(cls, proto_rpc_def, proto_req_def, proto_res_def)

    defer(do_compile)

def compile_remoteclass_del(cls: Type[T], del_method: Callable[P, R]) -> None:
    define_remoteclass_instance_ref(cls)

    def do_compile() -> None:
        typed_params = get_method_typed_params(cls, del_method)

        proto_rpc_name = names.proto_method_rpc_name(cls, del_method)
        proto_req_name = names.proto_method_req_name(cls, del_method)
        proto_res_name = names.proto_method_res_name(cls, del_method)

        proto_rpc_def = f"  rpc {proto_rpc_name} ({proto_req_name}) returns ({proto_res_name});"
        proto_req_def = gen_proto_msg_def(proto_req_name, typed_params)
        proto_res_def = gen_proto_msg_def(proto_res_name, {})

        add_remoteclass_method_def(cls, proto_rpc_def, proto_req_def, proto_res_def)

    defer(do_compile)

def compile_remoteclass_method(cls: Type[T], method: Callable[P, R]) -> None:
    define_remoteclass_instance_ref(cls)

    def do_compile() -> None:
        typed_params = get_method_typed_params(cls, method)
        return_type = get_method_return_type(cls, method)

        proto_rpc_name = names.proto_method_rpc_name(cls, method)
        proto_req_name = names.proto_method_req_name(cls, method)
        proto_res_name = names.proto_method_res_name(cls, method)

        proto_rpc_def = f"  rpc {proto_rpc_name} ({proto_req_name}) returns ({proto_res_name});"
        proto_req_def = gen_proto_msg_def(proto_req_name, typed_params)
        proto_res_def = gen_proto_msg_def(proto_res_name, {"return": return_type})

        add_remoteclass_method_def(cls, proto_rpc_def, proto_req_def, proto_res_def)

    defer(do_compile)

def compile_proto() -> Tuple[ModuleType, ModuleType]:
    proto_primitive_type_mapping.update(get_primitive_type_registry())
    do_deferred()

    with tempfile.TemporaryDirectory(prefix="egrpc_") as tempdir:
        proto_file = os.path.join(tempdir, "dynamic.proto")

        with open(proto_file, "w") as f:
            f.write(proto_file_content())

        protoc.main(
            [
                "grpc_tools.protoc",
                f"--proto_path={tempdir}",
                f"--python_out={tempdir}",
                f"--grpc_python_out={tempdir}",
                proto_file,
            ]
        )

        def import_dynamic_module(module_name: str, filename: str) -> ModuleType:
            filepath = os.path.join(tempdir, filename)
            spec = importlib.util.spec_from_file_location(module_name, filepath)
            assert spec is not None
            assert spec.loader is not None
            module = importlib.util.module_from_spec(spec)
            sys.modules[module_name] = module
            spec.loader.exec_module(module)
            return module

        dynamic_pb2 = import_dynamic_module("dynamic_pb2", "dynamic_pb2.py")
        dynamic_pb2_grpc = import_dynamic_module("dynamic_pb2_grpc", "dynamic_pb2_grpc.py")

        return dynamic_pb2, dynamic_pb2_grpc

def proto_file_content() -> str:
    proto_primitive_type_mapping.update(get_primitive_type_registry())
    do_deferred()
    return proto_header + proto_content



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Union, List, Tuple, Dict, Optional, Iterator, Any, Generic, TypeVar, get_origin
import sys
import os
import multiprocessing
import time
import math
import gc
import weakref
import pytest
import egrpc

env_name = "client"

def serve(port: int, error_queue: Any) -> None:
    global env_name
    env_name = "server"

    # redirect stdout to parent process
    sys.stdout = os.fdopen(sys.stdout.fileno(), "w", buffering=1)

    try:
        egrpc.serve(port)
    except AssertionError as e:
        error_queue.put(str(e))
    except Exception as e:
        error_queue.put(f"Unexpected error: {e}")

@pytest.fixture(scope="module")
def server() -> Iterator[None]:
    port = 12345

    error_queue: Any = multiprocessing.Queue()

    server_process = multiprocessing.Process(target=serve, args=(port, error_queue))
    server_process.start()

    time.sleep(1)

    if not server_process.is_alive():
        raise RuntimeError("Server failed to start.")

    egrpc.connect("localhost", port)

    try:
        yield

    finally:
        while not error_queue.empty():
            error_message = error_queue.get()
            pytest.fail(f"Server process error: {error_message}")

        server_process.terminate()
        server_process.join()

        egrpc.disconnect()

@egrpc.function
def get_gvar() -> str:
    return env_name

def test_remote_exec(server: Any) -> None:
    assert env_name == "client"
    assert get_gvar() == "server"

@egrpc.function
def func1(name: str, age: int) -> str:
    return f"{name}: {age}"

@egrpc.function
def func2(name: str, weight: Union[int, float]) -> str:
    return f"{name}: {weight:.2f}"

@egrpc.function
def func3(x: Union[int, float]) -> Union[int, float]:
    return x * x

@egrpc.function
def func4(lst: Union[List[int], List[str]], n: int) -> Union[List[int], List[str]]:
    return lst * n

@egrpc.function
def func5(lst: List[Union[int, str]], n: int) -> List[Union[int, str]]:
    return lst * n

@egrpc.function
def func6(tup: Tuple[int, str]) -> Tuple[str, int]:
    return tup[1], tup[0]

@egrpc.function
def func7(d: Dict[str, int]) -> Dict[str, List[str]]:
    return {k: [k] * v for k, v in d.items()}

@egrpc.function
def func8(x: Optional[int] = None) -> Optional[int]:
    return x * x if x is not None else None

@egrpc.function
def func9(x: Tuple[int, ...]) -> Tuple[float, ...]:
    return tuple(float(n) for n in x)

@egrpc.function
def func_slice(s: slice) -> slice:
    return s

EllipsisType = type(...)

@egrpc.function
def func_ellipsis(e: EllipsisType) -> EllipsisType:  # type: ignore[valid-type]
    return e

@egrpc.function
def func_slice_tuple(idx: Tuple[Union[int, slice], ...]) -> Tuple[Union[int, slice], ...]:
    return idx

def test_function(server: Any) -> None:
    assert func1("Alice", 30) == "Alice: 30"
    assert func2("Bob", 60) == "Bob: 60.00"
    assert func2("Bob", 60.2) == "Bob: 60.20"
    assert func3(2) == 4
    assert func3(4.3) == pytest.approx(4.3 * 4.3)
    assert func4([1, 2, 3], 2) == [1, 2, 3, 1, 2, 3]
    assert func4(["a", "b", "c"], 2) == ["a", "b", "c", "a", "b", "c"]
    assert func5([1, "b", 3], 2) == [1, "b", 3, 1, "b", 3]
    assert func6((1, "a")) == ("a", 1)
    assert func7({"a": 1, "b": 2}) == {"a": ["a"], "b": ["b", "b"]}
    assert func8(2) == 4
    assert func8() is None
    assert func9((1, 2, 3)) == (1.0, 2.0, 3.0)
    assert func_slice(slice(1, 10, 2)) == slice(1, 10, 2)
    assert func_slice(slice(None, 5)) == slice(None, 5)
    assert func_slice(slice(3, None)) == slice(3, None)
    assert func_slice(slice(None)) == slice(None)
    assert func_ellipsis(...) is ...
    assert func_slice_tuple((slice(None), slice(None))) == (slice(None), slice(None))

@egrpc.dataclass
class Data():
    x: int
    y: float

@egrpc.function
def dcfunc1(d: Data) -> float:
    return d.x * d.y

@egrpc.function
def dcfunc2(d1: Data, d2: Data) -> bool:
    return d1 == d2

def test_dataclass(server: Any) -> None:
    d = Data(3, 2.2)
    assert dcfunc1(d) == pytest.approx(3 * 2.2)
    assert dcfunc2(d, d) is True
    assert dcfunc2(d, Data(2, 2.2)) is False

@egrpc.remoteclass
class Point():
    @egrpc.method
    def __init__(self, x: Union[int, float], y: Union[int, float], z: Union[int, float]):
        self.x = x
        self.y = y
        self.z = z

    @egrpc.method
    def norm(self) -> float:
        return math.sqrt(self.x * self.x + self.y * self.y + self.z * self.z)

    @egrpc.method
    def __str__(self) -> str:
        return ",".join([str(self.x), str(self.y), str(self.z)])

@egrpc.function
def identity(p: Point) -> Point:
    return p

@egrpc.function
def norm(p: Point) -> float:
    return p.norm()

def test_remoteclass(server: Any) -> None:
    p1 = Point(1, 2, 3)
    p2 = Point(1.1, 2.2, 3.3)

    assert p1 != p2
    assert identity(p1) == p1
    assert p1.norm() == pytest.approx(math.sqrt(14))
    assert norm(p1) == pytest.approx(math.sqrt(14))
    assert str(p1) == "1,2,3"

    with pytest.raises(AttributeError):
        p1.x

@egrpc.remoteclass
class Hoge():
    def __init__(self, x: int):
        self.x = x
        self._fuga = Fuga(str(x))

    @egrpc.property
    def value(self) -> int:
        return self.x

    @value.setter
    def value(self, x: int) -> None:
        self.x = x
        self._fuga.value = str(x)

    @egrpc.method
    def fuga(self) -> "Fuga":
        return self._fuga

@egrpc.remoteclass
class Fuga():
    def __init__(self, x: str):
        self.x = x

    @egrpc.property
    def value(self) -> str:
        return self.x

    @value.setter
    def value(self, x: str) -> None:
        self.x = x

    @egrpc.method
    def hoge(self) -> "Hoge":
        return Hoge(int(self.x))

@egrpc.function
def gen_hoge(x: int) -> Hoge:
    return Hoge(x)

def test_remoteclass2(server: Any) -> None:
    hoge = gen_hoge(1)
    fuga = hoge.fuga()
    hoge_ = fuga.hoge()
    assert hoge.value == hoge_.value

    hoge.value = 2
    assert hoge.value == 2
    assert hoge.fuga().value == "2"

def test_remoteclass3(server: Any) -> None:
    hoge = gen_hoge(7)
    fuga1 = hoge.fuga()
    fuga2 = hoge.fuga()
    assert fuga1 is fuga2
    assert fuga1.value == "7"

    ref = weakref.ref(fuga1)
    del fuga1
    del fuga2
    gc.collect()
    assert ref() is None

    fuga3 = hoge.fuga()
    assert fuga3.value == "7"

@egrpc.remoteclass
class Animal:
    def __init__(self, name: str):
        self._name = name

    @egrpc.property
    def name(self) -> str:
        return self._name

    @egrpc.method
    def speak(self) -> str:
        return "..."

@egrpc.remoteclass
class Dog(Animal):
    def __init__(self, name: str):
        super().__init__(name)

    @egrpc.method
    def speak(self) -> str:
        return "woof"

@egrpc.remoteclass
class Cat(Animal):
    def __init__(self, name: str):
        super().__init__(name)

    @egrpc.method
    def speak(self) -> str:
        return "meow"

@egrpc.remoteclass
class Mammal(Animal):
    def __init__(self, name: str, legs: int):
        super().__init__(name)
        self._legs = legs

    @egrpc.property
    def legs(self) -> int:
        return self._legs

@egrpc.remoteclass
class Horse(Mammal):
    def __init__(self, name: str):
        super().__init__(name, 4)

    @egrpc.method
    def speak(self) -> str:
        return "neigh"

@egrpc.function
def create_animal(kind: str, name: str) -> Animal:
    if kind == "dog":
        return Dog(name)
    elif kind == "cat":
        return Cat(name)
    elif kind == "horse":
        return Horse(name)
    else:
        return Animal(name)

@egrpc.function
def get_animal_name(animal: Animal) -> str:
    return animal.name

def test_inheritance(server: Any) -> None:
    dog = create_animal("dog", "Pochi")
    assert isinstance(dog, Dog)
    assert dog.name == "Pochi"
    assert dog.speak() == "woof"
    assert get_animal_name(dog) == "Pochi"

    cat = create_animal("cat", "Tama")
    assert isinstance(cat, Cat)
    assert cat.name == "Tama"
    assert cat.speak() == "meow"
    assert get_animal_name(cat) == "Tama"

    horse = create_animal("horse", "Uma")
    assert isinstance(horse, Horse)
    assert horse.name == "Uma"
    assert horse.speak() == "neigh"
    assert horse.legs == 4
    assert get_animal_name(horse) == "Uma"

    animal = create_animal("unknown", "X")
    assert isinstance(animal, Animal)
    assert animal.name == "X"
    assert animal.speak() == "..."

@egrpc.function
def raise_value_error(msg: str) -> None:
    raise ValueError(msg)

@egrpc.function
def raise_type_error(msg: str) -> None:
    raise TypeError(msg)

class CustomError(Exception):
    pass

@egrpc.function
def raise_custom_error(msg: str) -> None:
    raise CustomError(msg)

@egrpc.remoteclass
class ErrorThrower:
    @egrpc.method
    def __init__(self, value: int):
        self._value = value

    @egrpc.method
    def raise_if_negative(self) -> int:
        if self._value < 0:
            raise ValueError(f"value must be non-negative, got {self._value}")
        return self._value

def test_exception_propagation(server: Any) -> None:
    with pytest.raises(ValueError) as exc_info_ve:
        raise_value_error("test value error")
    assert str(exc_info_ve.value) == "test value error"

    with pytest.raises(TypeError) as exc_info_te:
        raise_type_error("test type error")
    assert str(exc_info_te.value) == "test type error"

    with pytest.raises(CustomError) as exc_info_ce:
        raise_custom_error("test custom error")
    assert str(exc_info_ce.value) == "test custom error"

def test_exception_propagation_method(server: Any) -> None:
    obj = ErrorThrower(10)
    assert obj.raise_if_negative() == 10

    obj_neg = ErrorThrower(-5)
    with pytest.raises(ValueError) as exc_info:
        obj_neg.raise_if_negative()
    assert "value must be non-negative" in str(exc_info.value)

_T = TypeVar("_T")

class _MyInt(int, Generic[_T]):
    pass

egrpc.register_primitive_type(_MyInt, "int64")

def _my_int_subtype_rule(t1: Any, t2: Any) -> Optional[bool]:
    o1 = get_origin(t1)
    o2 = get_origin(t2)
    if o1 is _MyInt and o2 is None:
        return t2 in (int, float)
    if o1 is None and o2 is _MyInt:
        return False
    return None

egrpc.register_subtype_rule(_my_int_subtype_rule)

@egrpc.function
def my_int_identity(x: _MyInt[Any]) -> _MyInt[Any]:
    return x

@egrpc.function
def accept_my_int_or_float(x: Union[_MyInt[Any], float]) -> float:
    return float(x)

def test_custom_type_registration(server: Any) -> None:
    assert my_int_identity(_MyInt(42)) == 42
    assert accept_my_int_or_float(_MyInt(7)) == 7.0
    assert accept_my_int_or_float(3.14) == pytest.approx(3.14)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TypeVar, Type, Any, Dict, get_args
import weakref

from .util import TypeHint

T = TypeVar("T")

WeakRef = weakref.ref

InstanceRefType = int

def init_remoteclass(cls: Type[T]) -> None:
    if not hasattr(cls, "__del__"):
        def __del__(self: Any) -> None:
            pass
        setattr(cls, "__del__", __del__)

    setattr(cls, "__instance_count", 0)
    setattr(cls, "__instance_map_server", {})
    setattr(cls, "__instance_map_client", {})

def _get_instance_map_server(cls: Type[T]) -> Dict[InstanceRefType, T]:
    instance_map: Dict[InstanceRefType, T] = getattr(cls, "__instance_map_server")
    return instance_map

def _get_instance_map_client(cls: Type[T]) -> Dict[InstanceRefType, WeakRef[T]]:
    instance_map: Dict[InstanceRefType, WeakRef[T]] = getattr(cls, "__instance_map_client")
    return instance_map

def _get_instance_ref(obj: T) -> int:
    instance_ref: int = getattr(obj, "__instance_ref")
    return instance_ref

def get_ref_from_instance(cls: Type[T], obj: T, on_server: bool) -> InstanceRefType:
    if on_server and not hasattr(obj, "__instance_ref"):
        instance_ref: int = getattr(cls, "__instance_count")
        setattr(cls, "__instance_count", instance_ref + 1)
        setattr(obj, "__instance_ref", instance_ref)
    else:
        instance_ref = _get_instance_ref(obj)

    if on_server:
        instance_map_server = _get_instance_map_server(cls)
        instance_map_server[instance_ref] = obj

    return instance_ref

def get_instance_from_ref(cls: Type[T], instance_ref: InstanceRefType, type_hint: TypeHint, on_server: bool) -> T:
    if on_server:
        instance_map_server = _get_instance_map_server(cls)
        if instance_ref not in instance_map_server:
            raise ValueError(f"Unknown instance ref {instance_ref} for {cls.__name__}")
        return instance_map_server[instance_ref]

    else:
        instance_map_client = _get_instance_map_client(cls)

        if instance_ref in instance_map_client:
            obj = instance_map_client[instance_ref]()
            if obj is not None:
                return obj
            # Weakref target was GC'd; re-create the proxy instance.
            del instance_map_client[instance_ref]

        obj = object.__new__(cls)

        # If this type is a generic type, set `__orig_class__` attribute
        # so that this information is used for type-based dispatch
        if len(get_args(type_hint)) > 0:
            setattr(obj, "__orig_class__", type_hint)

        assign_ref_to_instance(cls, obj, instance_ref)
        return obj

def assign_ref_to_instance(cls: Type[T], obj: T, instance_ref: InstanceRefType) -> None:
    setattr(obj, "__instance_ref", instance_ref)
    # use weakref so that obj is garbage collected regardless of instance map
    instance_map_client = _get_instance_map_client(cls)
    assert instance_ref not in instance_map_client
    instance_map_client[instance_ref] = weakref.ref(obj)

def del_instance(cls: Type[T], obj: T) -> None:
    instance_map = _get_instance_map_server(cls)
    instance_ref = _get_instance_ref(obj)
    if instance_ref in instance_map:
        del instance_map[instance_ref]



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Any, Protocol, Type, Optional, Dict
from typing_extensions import Self

from .util import TypeHint

class PayloadType(Protocol):
    @classmethod
    def pack(cls, value: Any) -> Self:
        ...

    def unpack(self) -> Any:
        ...

_registry: Dict[Type[Any], Type[PayloadType]] = {}

def register_type(python_type: Type[Any], payload_cls: Type[PayloadType]) -> None:
    global _registry
    _registry[python_type] = payload_cls

def get_handler_for_type(type_hint: TypeHint) -> Optional[Type[PayloadType]]:
    global _registry
    for python_type, handler in _registry.items():
        if isinstance(type_hint, type) and issubclass(type_hint, python_type):
            return handler
    return None

_primitive_type_registry: Dict[type, str] = {}

def register_primitive_type(python_type: type, proto_type: str) -> None:
    global _primitive_type_registry
    _primitive_type_registry[python_type] = proto_type

def get_primitive_type_registry() -> Dict[type, str]:
    global _primitive_type_registry
    return _primitive_type_registry



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TypeVar, Callable, Type, Any, Optional, cast, TYPE_CHECKING
from typing_extensions import ParamSpec, dataclass_transform
import functools
import dataclasses

from .compiler import compile_function, compile_dataclass, compile_remoteclass, compile_remoteclass_init, compile_remoteclass_method, compile_remoteclass_del
from .proto_interface import pack_proto_function_request, pack_proto_function_response, unpack_proto_function_request, unpack_proto_function_response, pack_proto_method_request, pack_proto_method_response, unpack_proto_method_request, unpack_proto_method_response, ProtoMsg
from .grpc_interface import grpc_register_function, grpc_register_method, grpc_function_call, grpc_method_call, remote_connected
from .instance_ref import init_remoteclass, assign_ref_to_instance, del_instance

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")

def function_decorator(func: Callable[P, R]) -> Callable[P, R]:
    compile_function(func)

    def function_handler(self: Any, proto_req: ProtoMsg, context: Any) -> ProtoMsg:
        args = unpack_proto_function_request(func, proto_req)
        result = func(**args) # type: ignore[call-arg]
        return pack_proto_function_response(func, result)

    grpc_register_function(func, function_handler)

    @functools.wraps(func)
    def wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        if remote_connected():
            proto_req = pack_proto_function_request(func, *args, **kwargs)
            proto_res = grpc_function_call(func, proto_req)
            return cast(R, unpack_proto_function_response(func, proto_res))
        else:
            return func(*args, **kwargs)

    return wrapper

@dataclass_transform()
def dataclass_decorator(cls: Type[T]) -> Type[T]:
    datacls = dataclasses.dataclass(cls)
    compile_dataclass(datacls)
    return datacls

def method_decorator(method: Callable[P, R]) -> Callable[P, R]:
    setattr(method, "__egrpc_enabled", True)
    return method

# https://github.com/python/mypy/issues/6158
if TYPE_CHECKING:
    property_decorator = property
else:
    class property_decorator(property):
        def __init__(self,
                     fget: Callable[[Any], R],
                     fset: Optional[Callable[[Any, R], None]] = None,
                     fdel: Optional[Callable[[Any], None]]    = None,
                     fdoc: Optional[str]                      = None):
            setattr(self, "__egrpc_enabled", True)
            super().__init__(fget=fget, fset=fset)

def remoteclass_decorator(cls: Type[T]) -> Type[T]:
    init_remoteclass(cls)
    methods = {k: v for c in reversed(cls.__mro__) for k, v in c.__dict__.items() if hasattr(v, "__egrpc_enabled")}

    for method_name, method in methods.items():
        if isinstance(method, property_decorator):
            property_wrapper = register_remoteclass_property(cls, method)
            setattr(cls, method_name, property_wrapper)

        elif method_name == "__init__":
            init_wrapper = register_remoteclass_init(cls)
            setattr(cls, "__init__", init_wrapper)

        else:
            method_wrapper = register_remoteclass_method(cls, method)
            setattr(cls, method_name, method_wrapper)

    del_wrapper = register_remoteclass_del(cls)
    setattr(cls, "__del__", del_wrapper)

    compile_remoteclass(cls)

    return cls

def register_remoteclass_init(cls: Type[T]) -> Callable[..., None]:
    init_method = getattr(cls, "__init__")

    compile_remoteclass_init(cls, init_method)

    def init_handler(self: Any, proto_req: ProtoMsg, context: Any) -> ProtoMsg:
        args = unpack_proto_method_request(cls, init_method, proto_req)
        obj = cls(**args)
        return pack_proto_method_response(cls, init_method, obj)

    grpc_register_method(cls, init_method, init_handler)

    @functools.wraps(init_method)
    def init_wrapper(*args: Any, init_method: Callable[P, R] = init_method, **kwargs: Any) -> None:
        if remote_connected():
            proto_req = pack_proto_method_request(cls, init_method, *args, **kwargs)
            proto_res = grpc_method_call(cls, init_method, proto_req)
            instance_ref = unpack_proto_method_response(cls, init_method, proto_res)
            assign_ref_to_instance(cls, args[0], instance_ref)
        else:
            init_method(*args, **kwargs)

    return init_wrapper

def register_remoteclass_del(cls: Type[T]) -> Callable[..., None]:
    del_method = getattr(cls, "__del__")

    compile_remoteclass_del(cls, del_method)

    def del_handler(self: Any, proto_req: ProtoMsg, context: Any) -> ProtoMsg:
        args = unpack_proto_method_request(cls, del_method, proto_req)
        del_instance(cls, list(args.values())[0])
        return pack_proto_method_response(cls, del_method, None)

    grpc_register_method(cls, del_method, del_handler)

    @functools.wraps(del_method)
    def del_wrapper(self: Any) -> None:
        try:
            if remote_connected():
                proto_req = pack_proto_method_request(cls, del_method, self)
                grpc_method_call(cls, del_method, proto_req)
            else:
                del_method(self)
        except Exception:
            pass

    return del_wrapper

def register_remoteclass_method(cls: Type[T], method: Callable[P, R]) -> Callable[P, R]:
    compile_remoteclass_method(cls, method)

    def method_handler(self: Any, proto_req: ProtoMsg, context: Any) -> ProtoMsg:
        args = unpack_proto_method_request(cls, method, proto_req)
        result = method(**args) # type: ignore[call-arg]
        return pack_proto_method_response(cls, method, result)

    grpc_register_method(cls, method, method_handler)

    @functools.wraps(method)
    def method_wrapper(*args: P.args, **kwargs: P.kwargs) -> R:
        if remote_connected():
            proto_req = pack_proto_method_request(cls, method, *args, **kwargs)
            proto_res = grpc_method_call(cls, method, proto_req)
            return cast(R, unpack_proto_method_response(cls, method, proto_res))
        else:
            return method(*args, **kwargs)

    return method_wrapper

def register_remoteclass_property(cls: Type[T], prop: property_decorator) -> property:
    assert prop.fget is not None
    qualname = prop.fget.__qualname__

    if not prop.fget.__qualname__.endswith(".getter"):
        prop.fget.__qualname__ = f"{qualname}.getter"
    fget_wrapper = register_remoteclass_method(cls, prop.fget)

    if prop.fset is not None:
        if not prop.fset.__qualname__.endswith(".setter"):
            prop.fset.__qualname__ = f"{qualname}.setter"
        fset_wrapper = register_remoteclass_method(cls, prop.fset)
    else:
        fset_wrapper = None

    return property(fget=fget_wrapper, fset=fset_wrapper)



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from .entry_point import serve, connect, disconnect
from .registry import register_type, register_primitive_type
from .util import register_subtype_rule
from .compiler import proto_file_content
from .decorator import function_decorator as function, dataclass_decorator as dataclass, remoteclass_decorator as remoteclass, method_decorator as method, property_decorator as property
from .exception import RemoteError

__all__ = [
    "serve",
    "connect",
    "disconnect",
    "register_type",
    "register_primitive_type",
    "register_subtype_rule",
    "proto_file_content",
    "function",
    "dataclass",
    "remoteclass",
    "method",
    "property",
    "RemoteError",
]



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Optional
import os

from .compiler import compile_proto
from .proto_interface import init_proto
from .grpc_interface import init_grpc, init_server, init_client, del_client

def init() -> None:
    dynamic_pb2, dynamic_pb2_grpc = compile_proto()
    init_proto(dynamic_pb2)
    init_grpc(dynamic_pb2_grpc)

def serve(port: int, host: Optional[str] = None) -> None:
    init()
    server = init_server(port, host)
    server.start()
    bind_target = f"{host}:{port}" if host is not None else f"port {port}"
    print(f"Server started on {bind_target} (pid = {os.getpid()}).")
    server.wait_for_termination()

def connect(hostname: str, port: int) -> None:
    init()
    init_client(hostname, port)
    print(f"Client connected to server {hostname}:{port} (pid = {os.getpid()}).")

def disconnect() -> None:
    del_client()



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import Optional, Type, Tuple, List
import json
import builtins
import importlib

class RemoteError(Exception):
    pass

def serialize_exception(exc: Exception) -> Tuple[str, str]:
    info = {
        "type": f"{type(exc).__module__}.{type(exc).__name__}",
        "message": str(exc),
    }
    return ("x-exception-info", json.dumps(info))

def deserialize_exception(metadata: List[Tuple[str, str]]) -> Exception:
    metadata_dict = dict(metadata)
    info_json = metadata_dict.get("x-exception-info")
    if info_json is None:
        return RemoteError("Unknown remote error")

    info = json.loads(info_json)
    exc_cls = _get_exception_class(info["type"])
    return exc_cls(info["message"])

def _get_exception_class(type_name: str) -> Type[Exception]:
    parts = type_name.rsplit(".", 1)
    if len(parts) == 2:
        module_name, class_name = parts
        try:
            module = importlib.import_module(module_name)
            cls: Optional[Type[Exception]] = getattr(module, class_name, None)
            if cls is not None and isinstance(cls, type) and issubclass(cls, Exception):
                return cls
        except (ImportError, AttributeError):
            pass

    # Fallback: try builtins (ValueError, TypeError, etc.)
    class_name = type_name.split(".")[-1]
    builtin_cls: Optional[Type[Exception]] = getattr(builtins, class_name, None)
    if builtin_cls is not None and isinstance(builtin_cls, type) and issubclass(builtin_cls, Exception):
        return builtin_cls

    return RemoteError



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import get_type_hints, get_origin, get_args, Union, List, Tuple, Dict, Any, TypeVar, Callable, Type, Optional
from typing_extensions import ParamSpec
from collections.abc import Sequence, Mapping
import inspect

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")

TypeHint = Any

SubtypeRule = Callable[[TypeHint, TypeHint], Optional[bool]]
_subtype_rules: List[SubtypeRule] = []

def register_subtype_rule(rule: SubtypeRule) -> None:
    global _subtype_rules
    _subtype_rules.append(rule)

def check_subtype_rules(type_hint1: TypeHint, type_hint2: TypeHint) -> Optional[bool]:
    global _subtype_rules
    for rule in _subtype_rules:
        result = rule(type_hint1, type_hint2)
        if result is not None:
            return result
    return None

# https://github.com/python/mypy/issues/15630
def my_get_origin(type_hint: Any) -> Any:
    return get_origin(type_hint)

def is_subtype(type_hint1: TypeHint, type_hint2: TypeHint) -> bool:
    type_origin1 = my_get_origin(type_hint1)
    type_origin2 = my_get_origin(type_hint2)
    type_args1 = get_args(type_hint1)
    type_args2 = get_args(type_hint2)

    if type_origin1 is None and type_origin2 is None:
        try:
            return issubclass(type_hint1, type_hint2)
        except TypeError:
            return type_hint1 is type_hint2

    elif type_origin1 is Union and type_origin2 is Union:
        return all(any(is_subtype(th1, th2) for th2 in type_args2) for th1 in type_args1)

    elif type_origin2 is Union:
        return any(is_subtype(type_hint1, th2) for th2 in type_args2)

    elif type_origin1 is Union:
        return all(is_subtype(th1, type_hint2) for th1 in type_args1)

    else:
        custom_result = check_subtype_rules(type_hint1, type_hint2)
        if custom_result is not None:
            return custom_result

        if type_origin1 == type_origin2:
            # TODO: seriously consider covariant, contravariant, Sequence, Mapping, ...
            return (
                len(type_args1) != len(type_args2)
                and all(is_subtype(th1, th2) for th1, th2 in zip(type_args1, type_args2))
            )

        return False

def is_type_match(obj: Any, type_hint: TypeHint) -> bool:
    type_origin = my_get_origin(type_hint)
    type_args = get_args(type_hint)

    if type_origin is None:
        return isinstance(obj, type_hint)

    elif type_origin is Union:
        return any(is_type_match(obj, th) for th in type_args)

    elif type_origin in (tuple, Tuple):
        if len(type_args) == 2 and type_args[1] is Ellipsis:
            return (
                isinstance(obj, tuple)
                and all(is_type_match(o, type_args[0]) for o in obj)
            )
        else:
            return (
                isinstance(obj, tuple)
                and len(obj) == len(type_args)
                and all(is_type_match(o, th) for o, th in zip(obj, type_args))
            )

    elif type_origin in (list, List, Sequence):
        return (
            isinstance(obj, list)
            and all(is_type_match(o, type_args[0]) for o in obj)
        )

    elif type_origin in (dict, Dict, Mapping):
        return (
            isinstance(obj, dict)
            and all(is_type_match(k, type_args[0]) for k in obj.keys())
            and all(is_type_match(v, type_args[1]) for v in obj.values())
        )

    else:
        # Generic types: check isinstance against origin type
        try:
            if not isinstance(obj, type_origin):
                return False
        except TypeError:
            return False
        # If object has __orig_class__, also verify type arguments
        if hasattr(obj, "__orig_class__"):
            return all(is_subtype(th1, th2) for th1, th2 in zip(get_args(obj.__orig_class__), type_args))
        return True

def get_function_typed_params(func: Callable[P, R]) -> Dict[str, TypeHint]:
    type_hints = get_type_hints(func)
    param_names = list(inspect.signature(func).parameters.keys())
    return {param_name: type_hints[param_name] for param_name in param_names}

def get_function_return_type(func: Callable[P, R]) -> TypeHint:
    type_hints = get_type_hints(func)
    return type_hints["return"]

def get_class_typed_members(cls: Type[T]) -> Dict[str, TypeHint]:
    return get_type_hints(cls)

def get_method_typed_params(cls: Type[T], method: Callable[P, R]) -> Dict[str, TypeHint]:
    type_hints = get_type_hints(method)
    param_names = list(inspect.signature(method).parameters.keys())
    return {param_names[0]: cls,
            **{param_name: type_hints[param_name] for param_name in param_names[1:]}}

def get_method_return_type(cls: Type[T], method: Callable[P, R]) -> TypeHint:
    type_hints = get_type_hints(method)
    return type_hints["return"]

def normalize_args(func: Callable[P, R], *args: P.args, **kwargs: P.kwargs) -> Dict[str, Any]:
    sig = inspect.signature(func)
    bound_args = sig.bind(*args, **kwargs)
    bound_args.apply_defaults()
    return bound_args.arguments



# Copyright 2025 TOYOTA MOTOR CORPORATION.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import annotations
from typing import TypeVar, Callable, Type, Any, Optional, NoReturn, Dict
from typing_extensions import ParamSpec
from types import ModuleType
from concurrent import futures
import traceback
import grpc # type: ignore[import-untyped]

from . import names
from .proto_interface import ProtoMsg
from .exception import serialize_exception, deserialize_exception

T = TypeVar("T")
P = ParamSpec("P")
R = TypeVar("R")

dynamic_pb2_grpc: Optional[ModuleType] = None

def init_grpc(module: ModuleType) -> None:
    global dynamic_pb2_grpc
    dynamic_pb2_grpc = module

def get_grpc_module() -> ModuleType:
    global dynamic_pb2_grpc
    assert dynamic_pb2_grpc is not None
    return dynamic_pb2_grpc

HandlerType = Callable[[Any, ProtoMsg, Any], ProtoMsg]

proto_handlers: Dict[str, Dict[str, HandlerType]] = {}

def _abort_with_exception(context: Any, exc: Exception) -> NoReturn:
    traceback.print_exc()
    key, value = serialize_exception(exc)
    context.set_trailing_metadata([(key, value)])
    context.abort(grpc.StatusCode.INTERNAL, str(exc))
    raise AssertionError("unreachable")  # context.abort() raises, but type checker doesn't know

def grpc_register_function(func: Callable[P, R], handler: HandlerType) -> None:
    proto_service_name = names.proto_function_service_name(func)
    proto_rpc_name = names.proto_function_rpc_name(func)

    def wrapper(self: Any, proto_req: ProtoMsg, context: Any) -> ProtoMsg:
        try:
            return handler(self, proto_req, context)
        except Exception as exc:
            _abort_with_exception(context, exc)

    global proto_handlers
    assert proto_service_name not in proto_handlers
    proto_handlers[proto_service_name] = {proto_rpc_name: wrapper}

def grpc_register_method(cls: Type[T], method: Callable[P, R], handler: HandlerType) -> None:
    proto_service_name = names.proto_remoteclass_service_name(cls)
    proto_rpc_name = names.proto_method_rpc_name(cls, method)

    def wrapper(self: Any, proto_req: ProtoMsg, context: Any) -> ProtoMsg:
        try:
            return handler(self, proto_req, context)
        except Exception as exc:
            _abort_with_exception(context, exc)

    global proto_handlers
    if proto_service_name not in proto_handlers:
        proto_handlers[proto_service_name] = {}
    proto_handlers[proto_service_name][proto_rpc_name] = wrapper

MAX_MESSAGE_LENGTH = 256 * 1024 * 1024  # 256MB

def init_server(port: int, host: Optional[str] = None) -> Any:
    server = grpc.server(futures.ThreadPoolExecutor(max_workers=1),
                         maximum_concurrent_rpcs=1,
                         options=(
                             ("grpc.so_reuseport", 0),
                             ("grpc.max_send_message_length", MAX_MESSAGE_LENGTH),
                             ("grpc.max_receive_message_length", MAX_MESSAGE_LENGTH),
                         ))

    global proto_handlers
    for proto_service_name, handlers in proto_handlers.items():
        DynamicServicer = type(f"{proto_service_name}DynamicServicer",
                               (getattr(dynamic_pb2_grpc, f"{proto_service_name}Servicer"),),
                               handlers)

        add_service_fn = getattr(dynamic_pb2_grpc, f"add_{proto_service_name}Servicer_to_server")
        add_service_fn(DynamicServicer(), server)

    bind_host = host if host is not None else "[::]"
    server.add_insecure_port(f"{bind_host}:{port}")

    return server

ChannelType = Any

client_channel: Optional[ChannelType] = None
client_stubs: Dict[str, Any] = {}

def init_client(hostname: str, port: int) -> None:
    global client_channel, client_stubs
    client_channel = grpc.insecure_channel(
        f"{hostname}:{port}",
        options=(
            ("grpc.max_send_message_length", MAX_MESSAGE_LENGTH),
            ("grpc.max_receive_message_length", MAX_MESSAGE_LENGTH),
        )
    )
    client_stubs = {}

def del_client() -> None:
    global client_channel, client_stubs
    client_channel = None
    client_stubs = {}

def get_client_channel() -> ChannelType:
    global client_channel
    assert client_channel is not None
    return client_channel

def _get_stub(proto_service_name: str) -> Any:
    stub = client_stubs.get(proto_service_name)
    if stub is None:
        stub = getattr(get_grpc_module(), f"{proto_service_name}Stub")(get_client_channel())
        client_stubs[proto_service_name] = stub
    return stub

def remote_connected() -> bool:
    return client_channel is not None

def grpc_function_call(func: Callable[P, R], proto_req: ProtoMsg) -> ProtoMsg:
    proto_service_name = names.proto_function_service_name(func)
    proto_rpc_name = names.proto_function_rpc_name(func)

    stub = _get_stub(proto_service_name)
    try:
        proto_res = getattr(stub, proto_rpc_name)(proto_req)
    except grpc.RpcError as e:
        raise deserialize_exception(e.trailing_metadata()) from None

    return proto_res

def grpc_method_call(cls: Type[T], method: Callable[P, R], proto_req: ProtoMsg) -> ProtoMsg:
    proto_service_name = names.proto_remoteclass_service_name(cls)
    proto_rpc_name = names.proto_method_rpc_name(cls, method)

    stub = _get_stub(proto_service_name)
    try:
        proto_res = getattr(stub, proto_rpc_name)(proto_req)
    except grpc.RpcError as e:
        raise deserialize_exception(e.trailing_metadata()) from None

    return proto_res
